# Local development and testing with CLI and Data Client Library

> **Objectives**: Understand how to implement and test a processing application with local catalogs.
>
> **Complexity**: Intermediate
>
> **Time to complete**: 45 min
>
> **Prerequisites**: [Use Spark Connector to read and write data](../spark-connector/README.md) (for the logic of the sample processing application)
>
> **Source code**: <a href="local-development-workflow.zip" download target="_blank">Download</a>

This example demonstrates how to do the following:

- Use the CLI to create a local output catalog to run your processing pipeline during development
- Use the CLI to create a local copy of the input platform hosted catalog, to create a fully offline
  development environment
- Use the Data Client Library, local catalogs, and your favorite testing framework to test your
  processing pipeline.

## About local catalogs

A local catalog is a special catalog encoded in a file and stored in the `~/.here/local/` directory.
The CLI and Data Client Library can spawn a local version of the Data API that reads and writes to
local catalogs. Every local catalog file has the filename `<catalog-id>.db`, and can be renamed (to
change the catalog ID), copied, and moved to other machines, to share local catalogs within or across
engineering teams. Catalogs can be added to the `~/.here/local/` directory and be immediately
used with the CLI and the Data Client Library.

Local catalogs are identified by the special HRN `hrn:local:data:::<catalog-id>`.

No authentication and no access to the external network are needed to use local catalogs, hence you
do not incur any additional cost for accessing local catalogs. Since they are contained in a local
machine, they are not subject to naming conflicts within your realm.

Additionally, local catalogs can be optionally created and stored in memory, rather than persisted
on disk, which makes them especially useful during development and automated testing.

> #### Note
> Local catalogs may only be used for development and testing purposes. Running production 
> use cases with local catalogs is not permitted.

## A sample processing application

Local catalogs can be used during the development of a wide range of applications, from standalone
to Spark or Flink based applications, and with versioned, stream, index, and volatile layers.

In this tutorial you use Data Client Library's Spark Connector, but the steps to run and test
this application with local catalog can be used in all other scenarios.

First, develop an application that for each topology partitions in
[HERE Map Content]({{book.filtered.portalUrlBase}}/data/{{book.filtered.mapContentCatalog}}), counts
the number of topology nodes in each level 14 sub-tile, encodes this information in level 12
GeoJSON tiles, and saves the result in a output versioned layer:

![Sample processing logic](application.png)

You use different shades of red based on the number of nodes, to create a "heatmap" representing
the density of intersections in the sub-tile:

![Final output](output.png)

## Set up the Maven project

Create the following folder structure for the project:

```txt
local-development-workflow
└── src
    ├── main
    |   ├── java
    |   ├── scala
    |   └── resources
    └── test
        ├── java
        ├── scala
        └── resources
```

You can do this with a single `bash` command:

```bash
mkdir -p local-development-workflow/src/{main,test}/{java,scala,resources}
```

The Maven POM file is similar to that in the
[Verify Maven Settings](../verify-m2/README.md) example, with the parent POM
and dependencies sections updated:

Parent POM:
```
[snippet](pom.xml#parent)
```

Dependencies:
```
[snippet](pom.xml#dependencies)
```

## Create a local output catalog

First you develop your pipeline, reading from HERE Map Content and writing to a local
catalog. Further on, you create a local copy of the HERE Map Content catalog to produce a
completely offline development environment.

Creating a local catalog with the CLI is very similar to creating a catalog on the platform. You
are using the following catalog configuration:

```json
[snippet](local-development-workflow-output-catalog.json)
```

The catalog contains a GeoJSON layer at level 12 (the same as in the input catalog). Create the
catalog by running the following command:

```bash
olp local catalog create node-count "Node Count" \
  --config local-development-workflow-output-catalog.json
```

> #### Note
> If a billing tag is required in your realm, update the config file by adding the `billingTags: ["YOUR_BILLING_TAG"]` property to the `layer` section.

The CLI returns as follows:

```
Catalog hrn:local:data:::node-count has been created.
```

This means you have now access to the local catalog `hrn:local:data:::node-count`, and you can play
with the local variants of the CLI catalog commands:

```bash
olp local catalog list
```

```bash
olp local catalog show hrn:local:data:::node-count
```

```bash
olp local catalog layer show hrn:local:data:::node-count node-count
```

For more details on the available CLI commands for local catalogs, see the
[{{book.standard.cli}} documentation]({{book.filtered.devGuidesUrlBase}}/open-location-platform-cli/user_guide/index.html).

The catalog data is stored in the `~/.here/local/node-count.db` file, which you can verify with:

```bash
ls ~/.here/local/
```

You can now create a `pipeline-config.conf` file containing the input catalog (HERE Map Content) and
your newly created local catalog is the output catalog:

{% if book.product == 'olp' %}
```json
[snippet](pipeline-config.conf)
```
{% else %}
```json
[snippet](pipeline-config-cn.conf)
```
{% endif %}

## Implement the application

The sample application reads and processes all input partitions within a bounding box, specified in
the `application.conf` configuration file:

{% if book.product == 'olp' %}
```json
[snippet](application.conf)
```
{% else %}
```json
[snippet](application-cn.conf)
```
{% endif %}

To implement the logic, use the
[Data Client Library Spark connector]({{book.filtered.developerPortalUrl}}/documentation/data-client-library/dev_guide/client/spark-support-overview.html)
and typical Spark transformations. Notice that you do not need to fully understand the processing
logic to proceed with the tutorial. At the end of this exercise, test this code as a black
box. 

{% codetabs name="Scala", type="scala" -%}
{% raw %}
[snippet](src/main/scala/SparkConnectorLocalScala.scala)
{% endraw %}
{% language name="Java", type="java" -%}
{% raw %}
[snippet](src/main/java/SparkConnectorLocal.java)
{% endraw %}
{%- endcodetabs %}

## Compile and run locally

To run the application locally, execute the following command:

{% if book.product == 'olp' %}

{% codetabs name="Scala", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocalScala \
    -Dpipeline-config.file=pipeline-config.conf \
    -Dconfig.file=application.conf \
    -Dspark.master="local[*]"
{% endraw %}    
{% language name="Java", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocal \
    -Dpipeline-config.file=pipeline-config.conf \
    -Dconfig.file=application.conf \
    -Dspark.master="local[*]"
{% endraw %}
{%- endcodetabs %}

{% else %}

{% codetabs name="Scala", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocalScala \
    -Dpipeline-config.file=pipeline-config-cn.conf \
    -Dconfig.file=application-cn.conf \
    -Dspark.master="local[*]"
{% endraw %}    
{% language name="Java", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocal \
    -Dpipeline-config.file=pipeline-config-cn.conf \
    -Dconfig.file=application-cn.conf \
    -Dspark.master="local[*]"
{% endraw %}
{%- endcodetabs %}

{% endif %}

{% if book.product == 'olp' %}
After one run, you can inspect the local output layer with the CLI and the included local Data
Inspector:

```bash
olp local catalog layer inspect hrn:local:data:::node-count node-count
```

{% else %}

After one run, you can download with the CLI the GeoJSON partitions that have been published:

```bash
olp local catalog layer partition get hrn:local:data:::node-count node-count \
  --all --output local-geojson-partitions
```

{% endif %}

## Create an offline development environment

So far, you have only used an input catalog hosted on the platform. To create a development
environment which does not require access to the external network nor HERE credentials, you now
create a local copy of the HERE Map Content catalog, by running the following command:

{% if book.product == 'olp' %}
```bash
olp local catalog copy create hrn:here:data::olp-here:rib-2 \
  --id here-map-content \
  --include topology-geometry \
  --filter [heretile]=bounding-box:52.67551,52.33826,13.76116,13.08835
```
{% else %}
```bash
olp local catalog copy create hrn:here-cn:data::olp-cn-here:here-map-content-china-2 \
  --id here-map-content \
  --include topology-geometry \
  --filter [heretile]=bounding-box:41.0596,39.4416,117.508,115.4172
```
{% endif %}

This creates a local catalog `hrn:local:data:::here-map-content`. Copy the latest version of
of the [HERE Map Content catalog]({{book.filtered.portalUrlBase}}/data/{{book.filtered.mapContentCatalog}}), and initialize and copy only the `topology-geometry` layer. Include only the partitions contained in the specified bounding box. Local catalog copies are special local catalogs that retain information about the source catalog, which allows the local copy to be updated by running:

```bash
olp local catalog copy update hrn:local:data:::here-map-content
``` 

You can now modify the pipeline configuration to use this local copy as input:

```json
[snippet](pipeline-config-local.conf)
```

Optionally, you do not need to access any catalog on the platform, you can
configure the Data Client Library to only use local catalogs, by setting
`here.platform.data-client.endpoint-locator.discovery-service-env=local`. This prevents the
library from trying to fetch an authentication token:

{% if book.product == 'olp' %}

{% codetabs name="Scala", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocalScala \
    -Dpipeline-config.file=pipeline-config-local.conf \
    -Dconfig.file=application.conf \
    -Dspark.master="local[*]" \
    -Dhere.platform.data-client.endpoint-locator.discovery-service-env=local
{% endraw %}    
{% language name="Java", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocal \
    -Dpipeline-config.file=pipeline-config-local.conf \
    -Dconfig.file=application.conf \
    -Dspark.master="local[*]" \
    -Dhere.platform.data-client.endpoint-locator.discovery-service-env=local
{% endraw %}
{%- endcodetabs %}

{% else %}

{% codetabs name="Scala", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocalScala \
    -Dpipeline-config.file=pipeline-config-local.conf \
    -Dconfig.file=application-cn.conf \
    -Dspark.master="local[*]" \
    -Dhere.platform.data-client.endpoint-locator.discovery-service-env=local
{% endraw %}    
{% language name="Java", type="sh" -%}
{% raw %}
mvn compile exec:java \
    -Dexec.mainClass=SparkConnectorLocal \
    -Dpipeline-config.file=pipeline-config-local.conf \
    -Dconfig.file=application-cn.conf \
    -Dspark.master="local[*]" \
    -Dhere.platform.data-client.endpoint-locator.discovery-service-env=local
{% endraw %}
{%- endcodetabs %}

{% endif %}

Now you can develop and run your application with no internet connection, for example, while
commuting to work.

> #### Note
> In this tutorial, you have used the query functionality of the Data Client Library to only process
> a small bounding box, therefore run times are fast regardless of whether you read from the
> real platform hosted catalog or from a local copy. For applications which process the
> entire input, it is recommended that you copy a clip of the input catalog first, to keep
> development cycles fast, and test your logic with the real data in a
> [Pipeline]({{book.filtered.devGuidesUrlBase}}/pipeline/index.html) after ensuring the
> logic works correctly with local data.

## Test the sample application

Local catalogs make it convenient to test the overall business logic of a processing application.
Unit testing single methods is an important but time consuming process, and testing the whole
pipeline as a black box is just as important. However, this usually requires creating and accessing catalog resources on the platform, which might be too slow and costly for frequently run automated test
suites. Each developer, or even each test run, needs to access a unique output catalog, which
further complicates the setup. Local _in-memory_ catalogs are ideal in this situation, as they are
kept in memory and are private to a single process instance.

> #### Note
> The maximum size of _in-memory_ catalogs is limited by the amount of memory available to the JVM.
> If you need to run performance or stress tests with large local catalogs, you may decide to use
> persisted catalogs (`here.platform.data-local.memory-mode=false`).

First create your test configuration, by creating the file `src/test/resources/application.conf`.
Configure the Data Client Library to use the `local` environment by default, which is needed to
create catalogs in that environment with the `AdminApi`, and enable _in-memory_ local catalogs:

```json
[snippet](src/test/resources/application.conf)
``` 

Run a single test to verify that the application correctly computes the number of nodes
in each level 14 tile. To this end, you first create some pseudo-random input data starting
from an expected output, encode it in topology partitions, run the pipeline, and verify that the
generated data matches the expected output. This setup can be used to run multiple tests, and with
different data, allowing to cover all edge-cases:

{% codetabs name="Scala", type="scala" -%}
{% raw %}
[snippet](src/test/scala/SparkConnectorLocalScalaTest.scala)
{% endraw %}
{% language name="Java", type="java" -%}
{% raw %}
[snippet](src/test/java/SparkConnectorLocalTest.java)
{% endraw %}
{%- endcodetabs %}

To run the test with Maven, execute this command:

```bash
mvn test -Dspark.master=local[*] 
```

## Deploy on the platform

Once you are satisfied with your application, you can build a fat JAR file as shown in the following:

```bash
mvn -Pplatform clean package
```

Deploy it via the Pipeline API, with no code modifications. For more details on pipeline
deployment, see the
[Pipelines Developer's Guide]({{book.filtered.devGuidesUrlBase}}/pipeline/index.html).

## Further information

For additional details on the topics covered in this tutorial, you can refer to the following sources:
- For information on the different layer types and configurations you can check the
  [Data API documentation]({{book.filtered.devGuidesUrlBase}}/data-api/data_dev_guide/index.html)
- To know more about the interactive querying and manipulation of catalogs, you should check the
  [{{book.standard.cli}} documentation]({{book.filtered.devGuidesUrlBase}}/open-location-platform-cli/user_guide/index.html) and the
  [Data section]({{book.filtered.portalUrlBase}}/data)
- For details on the Scala and Java APIs to access the platform, you can check the
  [Data Client Library Developer Guide]({{book.filtered.devGuidesUrlBase}}/data-client-library/dev_guide/index.html)
