/*
 * Copyright (c) 2018-2023 HERE Europe B.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.here.olp.util.geo.BoundingBox
import com.here.olp.util.quad.factory.HereQuadFactory
import com.here.platform.data.client.spark.LayerDataFrameReader.SparkSessionExt
import com.here.platform.data.client.spark.LayerDataFrameWriter.DataFrameExt
import com.here.platform.data.client.spark.scaladsl.{
  GroupedData,
  VersionedDataConverter,
  VersionedRowMetadata
}
import com.here.platform.pipeline.PipelineContext
import com.here.schema.rib.v2.topology_geometry_partition.TopologyGeometryPartition
import com.typesafe.config.ConfigBeanFactory
import org.apache.spark.sql.{Dataset, Row, SparkSession}
import scalapb.spark.Implicits._

object SparkConnectorLocalScala {

  // We expose two public methods, `main` and `run`. `main` reads the pipeline context and the
  // configuration from `pipeline-config.conf` and `application.conf`, while `run` accepts them
  // as parameters. This makes it convenient to test the logic later using `run` with different
  // configuration parameters, without having to pass them through the classpath.
  def main(args: Array[String]): Unit = {
    // Read the default pipeline context
    val pipelineContext = new PipelineContext

    // Read the bounding box configured in `application.conf`
    val bbox = ConfigBeanFactory.create(
      pipelineContext.applicationConfig.getConfig("tutorial.boundingBox"),
      classOf[BoundingBox]
    )

    run(pipelineContext, bbox)
  }

  def run(pipelineContext: PipelineContext, bbox: BoundingBox): Unit = {
    // Defines the input / output catalogs to be read / written to
    val inputHrn = pipelineContext.config.inputCatalogs("here-map-content")
    val outputHrn = pipelineContext.config.outputCatalog

    // Input / output layers
    val topologyGeometryLayer = "topology-geometry"
    val outputLayer = "node-count"

    val sparkSession: SparkSession =
      SparkSession.builder().appName("SparkTopologyNodeCount").getOrCreate()

    // Read the input data as a Dataset of topology partitions
    val topologyGeometryData: Dataset[TopologyGeometryPartition] = sparkSession
      .readLayer(inputHrn, topologyGeometryLayer)
      .query(
        s"mt_partition=inboundingbox=(${bbox.getNorth},${bbox.getSouth},${bbox.getEast},${bbox.getWest})"
      )
      .option("olp.connector.force-raw-data", value = true)
      .load()
      .select("data")
      .as[Array[Byte]]
      .map(TopologyGeometryPartition.parseFrom)

    // This is needed for implicit conversions - moved down to avoid conflicts with scalapb.spark.Implicits._
    import sparkSession.implicits._

    // Compute the output partitions
    val nodeCountPartitions: Dataset[NodeCountPartition] = topologyGeometryData.map { partition =>
      val quads = partition.node
        .flatMap(_.geometry)
        .map(
          g => HereQuadFactory.INSTANCE.getMapQuadByLocation(g.latitude, g.longitude, 14).getLongKey
        )
        .groupBy(identity)
        .map {
          case (tileId, seq) => NodeCount(tileId, seq.size)
        }

      NodeCountPartition(partition.partitionName, quads.toSeq)
    }

    // Encode the output partitions, convert to DataFrame and publish the output data
    nodeCountPartitions
      .map(p => (p.partitionName, p.toGeoJson.getBytes))
      .toDF("mt_partition", "data")
      .writeLayer(outputHrn, outputLayer)
      .withDataConverter(new VersionedDataConverter {
        def serializeGroup(rowMetadata: VersionedRowMetadata,
                           rows: Iterator[Row]): GroupedData[VersionedRowMetadata] =
          GroupedData(rowMetadata, rows.next().getAs[Array[Byte]]("data"))
      })
      .save()

    sparkSession.stop()
  }

  // Class representing the number of nodes in a level 14 sub-tile
  case class NodeCount(tileId: Long, count: Int) {
    def toGeoJson: String = {
      val box = HereQuadFactory.INSTANCE.getMapQuadByLongKey(tileId).getBoundingBox
      val e = box.getEast
      val w = box.getWest
      val n = box.getNorth
      val s = box.getSouth

      val color = {
        // from black (0 nodes) to red (500 nodes)
        s"rgb(${(count.min(500).toDouble / 500 * 255).toInt},0,0)"
      }

      """{"type":"Feature",""" +
        s""""geometry":{"type":"Polygon","coordinates":[[[$w,$s],[$e,$s],[$e,$n],[$w,$n],[$w,$s]]]},""" +
        s""""properties":{"description":{"tileId":$tileId,"count":$count},"style":{"color":"$color"}}}"""
    }
  }

  // Class representing the decoded content of an output partition
  case class NodeCountPartition(partitionName: String, counts: Seq[NodeCount]) {
    def toGeoJson: String =
      s"""{"type":"FeatureCollection","features":[${counts.map(_.toGeoJson).mkString(",")}]}"""
  }
}
