/*
 * Copyright (c) 2018-2023 HERE Europe B.V.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

import com.here.hrn.HRN
import com.here.olp.util.geo.BoundingBox
import com.here.olp.util.quad.factory.HereQuadFactory
import com.here.platform.data.client.engine.scaladsl.DataEngine
import com.here.platform.data.client.model._
import com.here.platform.data.client.scaladsl.{DataClient, NewPartition}
import com.here.platform.data.client.spark.DataClientSparkContextUtils.context._
import com.here.platform.data.client.spark.SparkSupport._
import com.here.platform.pipeline.{PipelineConfig, PipelineContext}
import com.here.schema.geometry.v2.geometry.Point
import com.here.schema.rib.v2.topology_geometry.Node
import com.here.schema.rib.v2.topology_geometry_partition.TopologyGeometryPartition
import org.junit.runner.RunWith
import org.scalatest.junit.JUnitRunner
import org.scalatest.{FlatSpec, Inspectors, Matchers}

import java.util.UUID
import scala.collection.JavaConverters._
import scala.util.Random

@RunWith(classOf[JUnitRunner])
class SparkConnectorLocalScalaTest extends FlatSpec with Matchers with Inspectors {

  // Specify the bounding box used by our pipeline under test.
  private val testBoundingBox = new BoundingBox(52.67551, 52.33826, 13.76116, 13.08835)

  // Utilities to create a random topology node within a bounding box.
  private val random = new Random(42)

  private def randomInRange(min: Double, max: Double): Double =
    min + (max - min) * random.nextDouble()

  private def randomNode(boundingBox: BoundingBox): Node =
    Node(
      geometry = Some(
        Point(
          latitude = randomInRange(boundingBox.getSouth, boundingBox.getNorth),
          longitude = randomInRange(boundingBox.getWest, boundingBox.getEast)
        )
      )
    )

  // Utility to create a catalog with a single HERE Tile versioned layer. The catalog will be local
  // because we use `here.platform.data-client.endpoint-locator.discovery-service-env=local` in
  // src/test/resources/application.conf. It will be kept in memory only, because we also use
  // `here.platform.data-local.memory-mode=true`.
  private def createCatalog(catalogIdPrefix: String, layerId: String): HRN =
    DataClient()
      .adminApi()
      .createCatalog(
        WritableCatalogConfiguration(
          catalogIdPrefix + "-" + UUID.randomUUID().toString,
          catalogIdPrefix,
          "summary",
          "description",
          Seq(
            WritableLayer(
              layerId,
              layerId,
              "summary",
              "description",
              LayerTypes.Versioned,
              HereTilePartitioning(List(12)),
              DurableVolume,
              contentType = "application/octet-stream"
            )
          )
        )
      )
      .awaitResult()

  // Utility to publish some test data in a catalog.
  private def writeData(hrn: HRN, layer: String, data: Map[String, Array[Byte]]): Unit = {
    val partitions = data.map {
      case (partitionId, data) =>
        NewPartition(
          partitionId,
          layer,
          NewPartition.ByteArrayData(data)
        )
    }

    DataEngine()
      .writeEngine(hrn)
      .publishBatch2(4, Some(Seq(layer)), Nil, partitions.iterator)
      .awaitResult()
  }

  // Utility to read the data from the latest version of a catalog.
  private def readData(hrn: HRN, layer: String): Map[String, Array[Byte]] = {
    val readEngine = DataEngine().readEngine(hrn)
    val queryApi = DataClient().queryApi(hrn)

    queryApi
      .getLatestVersion()
      .flatMap {
        case None => fail("Catalog is empty")
        case Some(v) => queryApi.getPartitionsAsIterator(v, layer)
      }
      .awaitResult()
      .map(p => p.partition -> readEngine.getDataAsBytes(p).awaitResult())
      .toMap
  }

  // Utility to create an input test catalog given the expected outcome. `expectedCounts` is a map
  // containing, for each level 14 tile ID, the number of nodes contained in that sub-tile.
  // This method creates the requested number of topology nodes in each level-14 tile, encodes them
  // in a topology partition and publishes them in a freshly created test catalog.
  private def createTestInputCatalog(expectedCounts: Map[Long, Int]): HRN = {
    val hrn = createCatalog("input", "topology-geometry")

    val partitions: Map[String, Array[Byte]] = expectedCounts
      .map {
        case (tileId, count) =>
          val boundingBox = HereQuadFactory.INSTANCE.getMapQuadByLongKey(tileId).getBoundingBox
          val nodes = (1L to count).map(_ => randomNode(boundingBox))
          tileId -> nodes
      }
      .groupBy {
        case (tileId, _) =>
          HereQuadFactory.INSTANCE.getMapQuadByLongKey(tileId).getAncestor(12).getLongKey
      }
      .map {
        case (tileId, nodes) =>
          tileId.toString -> TopologyGeometryPartition(
            partitionName = tileId.toString,
            node = nodes.values.reduce(_ ++ _)
          ).toByteArray
      }

    writeData(hrn, "topology-geometry", partitions)

    hrn
  }

  // Utility to create an output catalog for the pipeline.
  private def createTestOutputCatalog(): HRN = createCatalog("output", "node-count")

  // Utility to retrieve the actual node count per level 14 tile in the output catalog. It reads the
  // latest version of the catalog, retrieves the data, decodes it and returns the node count
  // encoded in each level 14 tile. This is used, after a run of the pipeline under test, to verify
  // that the node count is the expected on.
  private def readActualNodeCount(hrn: HRN): Map[Long, Int] = {

    // We use regular expressions to extract the tileId of the level 14 tiles and their node count.
    // Using a proper JSON parser is left as an exercise.
    val descriptionRegex =
      """"description":\s*\{"tileId":\s*(\d+),\s*"count":\s*(\d+)}""".r.unanchored

    readData(hrn, "node-count").flatMap {
      case (_, data) =>
        val geoJson = new String(data)
        descriptionRegex.findAllMatchIn(geoJson).map(m => m.group(1).toLong -> m.group(2).toInt)
    }
  }

  // Randomly generated expected result. For each level 14 tile in `testBoundingBox` we generate a
  // random non-negative node count. We filter out zeros because those are by design not encoded by
  // the pipeline under test.
  private val expectedNodeCount: Map[Long, Int] = {
    HereQuadFactory.INSTANCE
      .iterableBoundingBoxToMapQuad(testBoundingBox, 14)
      .asScala
      .map { quad =>
        quad.getLongKey -> (random.nextInt.abs % 100)
      }
      .filter(_._2 != 0)
      .toMap
  }

  "SparkConnectorLocalScala" should "correctly compute the number of nodes in each sub-tile" in {
    // Create the input catalog and publish the test data.
    val inputHrn = createTestInputCatalog(expectedNodeCount)
    // Create an empty output catalog.
    val outputHrn = createTestOutputCatalog()

    val pipelineContext =
      PipelineContext(PipelineConfig(outputHrn, Map("here-map-content" -> inputHrn)))

    // Run the pipeline under test.
    SparkConnectorLocalScala.run(pipelineContext, testBoundingBox)

    // Check the output data.
    readActualNodeCount(outputHrn) should contain theSameElementsAs expectedNodeCount
  }
}
